/*
 * You received this file as part of
 * FlexRay4Linux - FlexRay protocol stack for Linux
 * 
 * This file contains:
 *
 * - Definitions that are also needed in userspace
 *
 * Copyright:
 * 
 * - (C) 2009-2011 Michael Arndt <flexray4linux@scriptkiller.de>,
 *   Robotics Research Lab, University of Kaiserslautern
 *
 */

#ifndef __FLEXRAY_H
#define __FLEXRAY_H

#ifdef __KERNEL__

#include <linux/socket.h> // for sa_family_t
#include <linux/sockios.h>

#endif

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#ifndef __KERNEL__

#include <sys/socket.h>

typedef unsigned char u8;
typedef unsigned short int u16;
typedef unsigned int u32;

#endif

typedef u8 boolean;



/*
 * This file contains the stuff that is also used
 * in userspace.
 * 
 */


/* FlexRay protocols */
#define FR_RAW		1 /* RAW sockets */
#define FR_NPROTO	2 /* dimension marker */


/**
 * the sockaddr struct for FlexRay sockets
 */
struct sockaddr_fr {
	
	/** the address family, this is AF_FLEXRAY */
	sa_family_t fr_family;
	/** interface index of FlexRay device */
	int         fr_ifindex;

};


/* there are 16 device-private ioctls available, from
 * SIOCDEVPRIVATE to SIOCDEVPRIVATE+15
 * TODO: is their use deprecated?
 */
#define	SIOCDEVFLEXRAY	SIOCDEVPRIVATE

struct flexray_conf_ioctl
{
	/** see below */
	u8 cmd;
	
	/** pointer to additional data */
	void *data;
};

#define FLEXRAY_INIT	0x1
#define FLEXRAY_SET_LOW_LEVEL_CONFIG 0x2
#define FLEXRAY_START_COMM 0x3
#define FLEXRAY_STOP_COMM 0x4
#define FLEXRAY_LEAVE_CONF 0x5
#define FLEXRAY_ENTER_CONF 0x6
#define FLEXRAY_ADD_TX_BUF 0x7
#define FLEXRAY_ADD_RX_BUF 0x8
#define FLEXRAY_WAKEUP 0x9
#define FLEXRAY_GET_POC_STATE 0xA

////////////////////////////


/* forward declarations */
struct flexray4linux_low_level_config;
typedef struct flexray4linux_low_level_config flexray4linux_low_level_config;

struct flexray4linux_tx_buffer_config;
typedef struct flexray4linux_tx_buffer_config flexray4linux_tx_buffer_config;

struct flexray4linux_rx_buffer_config;
typedef struct flexray4linux_rx_buffer_config flexray4linux_rx_buffer_config;

struct flexray4linux_frame;
typedef struct flexray4linux_frame flexray4linux_frame;

/**
 * possible POC-states of the controller,
 * if you change something here, you *must* also change
 * the flexray4linux_poc_state_cleartext array below!
 */
enum {
	/** currently being configured by host */
	FLEXRAY4LINUX_POC_STATE_CONFIG,
	/** after power-on */
	FLEXRAY4LINUX_POC_STATE_DEFAULT_CONFIG,
	/** after severe errors */
	FLEXRAY4LINUX_POC_STATE_HALT,
	/** normal communication */
	FLEXRAY4LINUX_POC_STATE_NORMAL_ACTIVE,
	/** after some errors */
	FLEXRAY4LINUX_POC_STATE_NORMAL_PASSIVE,
	/** after being configured by host */
	FLEXRAY4LINUX_POC_STATE_READY,
	/** startup phase */
	FLEXRAY4LINUX_POC_STATE_STARTUP,

	FLEXRAY4LINUX_POC_STATE_WAKEUP,
	
	/** unknown POC-state, used when POC-state of the
	 * controller is not known by the driver
	 */
	FLEXRAY4LINUX_POC_STATE_UNKNOWN,
};
typedef u8 flexray4linux_poc_state;

/**
 * the same order as flexray4linux_poc_state but as
 * cleartext-strings for debugging or monitoring-purposes,
 * please read the note above before changing anything!
 * This is defined in the corresponding source file to this header,
 * please edit it there!
 */
extern const char* flexray4linux_poc_state_cleartext[];

/**
 * possible channels to operate on
 */
enum {
	FLEXRAY4LINUX_CHANNEL_A,
	FLEXRAY4LINUX_CHANNEL_B,
	FLEXRAY4LINUX_CHANNEL_AB,
	FLEXRAY4LINUX_NO_CHANNEL
};
typedef u8 flexray4linux_channel_type;

/**
 *  possible buffer types
 */
enum {
	/** single buffered */
	FLEXRAY4LINUX_BUFFER_TYPE_SINGLE,
	/** double buffered */
	FLEXRAY4LINUX_BUFFER_TYPE_DOUBLE
};
typedef u8 flexray4linux_buffer_type;

/**
 * possible transmission types
 */
enum {
	/** event driven transmission */
	FLEXRAY4LINUX_TRANSMISSION_TYPE_EVENT,
	/** state driven transmission */
	FLEXRAY4LINUX_TRANSMISSION_TYPE_STATE
};
typedef u8 flexray4linux_transmission_type;

/**
 * possible commit types
 */
enum {
	FLEXRAY4LINUX_TRANSMISSION_COMMIT_TYPE_STREAMING,
	FLEXRAY4LINUX_TRANSMISSION_COMMIT_TYPE_IMMEDIATE
};
typedef u8 flexray4linux_transmission_commit_type;


/**
 * This type contains the configuration of a
 * FlexRay transmit (TX) buffer.
 */
struct flexray4linux_tx_buffer_config {
	
	/** frame ID */
	u16 frame_id;
	
	/** the CRC for sending the header */
	u16 header_crc;

	/** payload length in 16 bit words (0..127) */
    u8 payload_length;
    
    /** message buffer buffering */
    flexray4linux_buffer_type mb_buffering;
    
    /** transmission mode */
    flexray4linux_transmission_type transmission_mode;
    
    /** transmission commit mode */
    flexray4linux_transmission_commit_type transmission_commit_mode;
    
    /** channel(s) to transmit on */
    flexray4linux_channel_type channel_enable;
    
    /** the payload preamble indicator */
    boolean payload_preamble;
    
    /** trasnmit cycle counter filter enable */
    boolean tx_cycle_counter_filter_enable;
    /** trasnmit cycle counter filter value */
    u8 tx_cycle_counter_filter_value;
    /** transmit cycle counter filter mask */
    u8 tx_cycle_counter_filter_mask;
    
} __attribute__((__packed__));

/**
 * This type contains the configuration of a
 * FlexRay receive (RX) buffer.
 */
struct flexray4linux_rx_buffer_config {

	/** frame ID*/
	u16 frame_id;

    /** channel(s) to receive from */
    flexray4linux_channel_type channel_enable;
    
    /** the payload preamble indicator */
    boolean payload_preamble;
    
    /** trasnmit cycle counter filter enable */
    boolean rx_cycle_counter_filter_enable;
    /** trasnmit cycle counter filter value */
    u8 rx_cycle_counter_filter_value;
    /** transmit cycle counter filter mask */
    u8 rx_cycle_counter_filter_mask;
    
} __attribute__((__packed__));

/**
 * this type contains the low level FlexRay configuration
 */
struct flexray4linux_low_level_config {

	u8 g_coldstart_attempts;
	u8 gd_action_points_offset;
	u8 gd_cas_rx_low_max;
	u8 gd_dynamic_slot_idle_phase;
	u8 gd_minislot;
	u8 gd_mini_slot_action_point_offset;
	u16 gd_static_slot;
	u8 gd_symbol_window;
	u8 gd_tss_transmitter;
	u8 gd_wakeup_symbol_rx_idle;
	u8 gd_wakeup_symbol_rx_low;
	u16 gd_wakeup_symbol_rx_window;
	u8 gd_wakeup_symbol_tx_idle;
	u8 gd_wakeup_symbol_tx_low;
	u8 g_listen_noise;
	u16 g_macro_per_cycle;
	u8 g_max_without_clock_correction_passive;
	u8 g_max_without_clock_correction_fatal;
	u16 g_number_of_minislots;
	u16 g_number_of_static_slots;
	u16 g_offset_correction_start;
	u8 g_payload_length_static;
	u8 g_sync_node_max;
	u8 g_network_management_vector_length;
	boolean p_allow_halt_due_to_clock;
	u8 p_allow_passive_to_active;
	flexray4linux_channel_type p_channels;
	u16 pd_accepted_startup_range;
	u8 p_cluster_drift_damping;
	u8 p_decoding_correction;
	u8 p_delay_compensation_cha;
	u8 p_delay_compensation_chb;
	u32 pd_listen_timeout;
	u16 pd_max_drift;
	u8 p_extern_offset_correction;
	u8 p_extern_rate_correction;
	u16 p_key_slot_id;
	boolean p_key_slot_used_for_startup;
	boolean p_key_slot_used_for_sync;
	u16 p_key_slot_header_crc;
	u16 p_latest_tx;
	u8 p_macro_initial_offset_a;
	u8 p_macro_initial_offset_b;
	u8 p_micro_initial_offset_a;
	u8 p_micro_initial_offset_b;
	u32 p_micro_per_cycle;
	u16 p_offset_correction_out;
	u16 p_rate_correction_out;
	boolean p_single_slot_enabled;
	flexray4linux_channel_type p_wakeup_channel;
	u8 p_wakeup_pattern;
	u8 p_micro_per_macro_nom;
	u8 p_payload_length_dyn_max;

} __attribute__((__packed__));

/**
 * This type describes a FlexRay frame.
 * Warning: before changing anything here read and understand the documentation
 * about flexray4linux_packet_data!
 */
struct flexray4linux_frame {
	
	/** 
	 * ID of the slot the FlexRay frame has been sent in
	 * or should be sent in
	 */
	u16 slot_id;
	
	/** number of the cycle the FlexRay frame has been sent in */
	u8 cycle;
	
	/** length of the data in 16bit words (not bytes!) */ // FIXME: these *are* bytes!!
	u8 length;
	
	/** the status of the slot the frame was been received in */
	u16 slot_status;
	
	/** the payload */
	u16 data[127]; /* this must be the last field in this struct */
	
} __attribute__((__packed__));



/* descriptions of slot status registers, see flexray4linux_frame */
/** Channel B: Frame valid */
#define FLEXRAY4LINUX_SLOT_STATUS_VFB	0x8000
/** Channel B: Frame is sync frame */
#define FLEXRAY4LINUX_SLOT_STATUS_SYB	0x4000
/** Channel B: Frame is null frame */
#define FLEXRAY4LINUX_SLOT_STATUS_NFB	0x2000
/** Channel B: Frame is startup frame */
#define FLEXRAY4LINUX_SLOT_STATUS_SUB	0x1000
/** Channel B: Syntax error */
#define FLEXRAY4LINUX_SLOT_STATUS_SEB	0x0800
/** Channel B: Content error */
#define FLEXRAY4LINUX_SLOT_STATUS_CEB	0x0400
/** Channel B: Boundary violation */
#define FLEXRAY4LINUX_SLOT_STATUS_BVB	0x0200
/** Channel B: Transmission conflict (only TX buffers) / Channel indicator */
#define FLEXRAY4LINUX_SLOT_STATUS_TCB	0x0100

/** Channel indicator / Channel B: Transmission conflict (only TX buffers) */
#define FLEXRAY4LINUX_SLOT_STATUS_CH	0x0100

/** Channel A: Frame valid */
#define FLEXRAY4LINUX_SLOT_STATUS_VFA	0x0080
/** Channel A: Frame is sync frame */
#define FLEXRAY4LINUX_SLOT_STATUS_SYA	0x0040
/** Channel A: Frame is null frame */
#define FLEXRAY4LINUX_SLOT_STATUS_NFA	0x0020
/** Channel A: Frame is startup frame */
#define FLEXRAY4LINUX_SLOT_STATUS_SUA	0x0010
/** Channel A: Syntax error */
#define FLEXRAY4LINUX_SLOT_STATUS_SEA	0x0008
/** Channel A: Content error */
#define FLEXRAY4LINUX_SLOT_STATUS_CEA	0x0004
/** Channel A: Boundary violation */
#define FLEXRAY4LINUX_SLOT_STATUS_BVA	0x0002
/** Channel A: Transmission conflict (only TX buffers) */
#define FLEXRAY4LINUX_SLOT_STATUS_TCA	0x0001

#endif
