/*
 * You received this file as part of
 * FlexRay4Linux - FlexRay protocol stack for Linux
 * 
 * This file contains:
 *
 * - Protoypes and definitions for FlexRay protocol modules using the
 *   PF_FLEXRAY core
 *
 * Copyright:
 * 
 * - (C) 2009-2011 Michael Arndt <flexray4linux@scriptkiller.de>,
 *   Robotics Research Lab, University of Kaiserslautern
 * - Heavily inspired by
 *   af_bluetooth from Bluez Copyright (C) 2000-2001 Qualcomm Incorporated and
 *   af_can Copyright (C) 2002-2007 Volkswagen Group Electronic Research
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY
 * CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS,
 * COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS
 * SOFTWARE IS DISCLAIMED.
 *
 * The provided data structures and external interfaces from this code
 * are not restricted to be used by modules with a GPL compatible license.
 * 
 */

/* FlexRay address family and sockets. */

#include <linux/module.h>

#include <linux/types.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/skbuff.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <net/sock.h>
#include <asm/ioctls.h>
#include <linux/kmod.h>
#include <linux/if_arp.h>

#include <linux/flexray.h>
#include "af_flexray.h"


#define VERSION "0.1"

MODULE_DESCRIPTION("FlexRay protocol stack for Linux: core");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Michael Arndt <flexray4linux@scriptkiller.de>");
MODULE_VERSION(VERSION);
MODULE_ALIAS_NETPROTO(PF_FLEXRAY);

// TODO: what exactly does this do?
static int stats_timer __read_mostly = 1;
module_param(stats_timer, int, S_IRUGO);
MODULE_PARM_DESC(stats_timer, "enable timer for statistics (default:on)");


/* list with AF_FLEXRAY sockets */
static struct hlist_head fr_sklist;
static DEFINE_RWLOCK(fr_sklist_lock);


//static struct kmem_cache *rcv_cache __read_mostly;

/* table of registered FlexRay protocols */
static struct fr_proto *proto_tab[FR_NPROTO] __read_mostly;
static DEFINE_SPINLOCK(proto_tab_lock);

// TODO: needed for proc
//struct timer_list fr_stattimer;   /* timer for statistics update */
//struct s_stats    fr_stats;       /* packet statistics */
//struct s_pstats   fr_pstats;      /* receive list statistics */

static int fr_insert_socket(struct hlist_head *list, struct sock *sk)
{
	write_lock_bh(&fr_sklist_lock);
	sk_add_node(sk, list);
	write_unlock_bh(&fr_sklist_lock);
	return 0;
}

static int fr_remove_socket(struct hlist_head *list, struct sock *sk)
{
	write_lock_bh(&fr_sklist_lock);
	sk_del_node_init(sk);
	write_unlock_bh(&fr_sklist_lock);
	return 0;
}

/*
 * AF_FLEXRAY socket functions
 */

static int fr_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	struct sock *sk = sock->sk;

	switch (cmd) {

	case SIOCGSTAMP:
		return sock_get_timestamp(sk, (struct timeval __user *)arg);

	default:
		return -ENOIOCTLCMD;
	}
}



/**
 * Register FlexRay transport protocol
 * 
 * @param p pointer to FlexRay protocol structure
 * @return
 *  0 on success
 *  -EINVAL invalid (out of range) protocol number
 *  -EBUSY  protocol already in use
 *  -ENOBUF if proto_register() fails
 */
int fr_proto_register(struct fr_proto *p)
{
	int proto = p->protocol;
	int err = 0;

	if (proto < 0 || proto >= FR_NPROTO) {
		printk(KERN_ERR "flexray: protocol number %d out of range\n",
		       proto);
		return -EINVAL;
	}

	err = proto_register(p->prot, 0);
	if (err < 0)
		return err;

	spin_lock(&proto_tab_lock);
	if (proto_tab[proto]) {
		printk(KERN_ERR "flexray: protocol %d already registered\n",
		       proto);
		err = -EBUSY;
	} else {
		proto_tab[proto] = p;

		/* use generic ioctl function if not defined by module */
		if (!p->ops->ioctl)
			p->ops->ioctl = fr_ioctl;

	}
	spin_unlock(&proto_tab_lock);

	if (err < 0)
		proto_unregister(p->prot);
	else
		printk(KERN_INFO "fleray: protocol %d successfully registered\n",
			proto);

	return err;
}
EXPORT_SYMBOL(fr_proto_register);

/**
 * ft_proto_unregister - unregister FlexRay transport protocol
 * @param p pointer to FlexRay protocol structure
 */
void fr_proto_unregister(struct fr_proto *p)
{
	int proto = p->protocol;

	spin_lock(&proto_tab_lock);
	if (!proto_tab[proto]) {
		printk(KERN_ERR "BUG: flexray: protocol %d is not registered\n",
		       proto);
	}
	proto_tab[proto] = NULL;
	spin_unlock(&proto_tab_lock);

	proto_unregister(p->prot);
	
	printk(KERN_INFO "flexray: protocol %d successfully unregistered\n",
			proto);
}
EXPORT_SYMBOL(fr_proto_unregister);

/**
 * Destroy the given FlexRay socket
 * TODO: ?
 * @param sk the socket
 */
static void fr_sock_destruct(struct sock *sk)
{
	printk(KERN_INFO "flexray: sock_destruct\n");
//	fr_remove_socket(&fr_sklist, sk);
	skb_queue_purge(&sk->sk_receive_queue);
}

/**
 * Create a FlexRay socket
 * @param net
 * @param sock
 * @param procotol
 * @param kern
 * @return
 */
static int fr_create(struct net *net, struct socket *sock, int protocol, int kern)
{
	struct sock *sk;
	struct fr_proto *p;
	int err = 0;

	printk(KERN_INFO "flexray: create\n");

	sock->state = SS_UNCONNECTED;

	if (protocol < 0 || protocol >= FR_NPROTO)
		return -EINVAL;

	if (net != &init_net)
		return -EAFNOSUPPORT;

#ifdef CONFIG_MODULES
	/* try to load protocol module if kernel is modular */
	if (!proto_tab[protocol]) {
		err = request_module("fr-proto-%d", protocol);

		/*
		 * In case of error we only print a message but don't
		 * return the error code immediately.  Below we will
		 * return -EPROTONOSUPPORT
		 */
		if (err && printk_ratelimit())
			printk(KERN_ERR "flexray: request_module "
			       "(fr-proto-%d) failed.\n", protocol);
	}
#endif

	spin_lock(&proto_tab_lock);
	p = proto_tab[protocol];
	if (p && !try_module_get(p->prot->owner))
		p = NULL;
	spin_unlock(&proto_tab_lock);

	/* check for available protocol and correct usage */

	if (!p)
		return -EPROTONOSUPPORT;

	if (p->type != sock->type) {
		err = -EPROTONOSUPPORT;
		goto errout;
	}

	if (p->capability >= 0 && !capable(p->capability)) {
		err = -EPERM;
		goto errout;
	}

	sock->ops = p->ops;

	sk = sk_alloc(net, PF_CAN, GFP_KERNEL, p->prot);
	if (!sk) {
		err = -ENOMEM;
		goto errout;
	}

	sock_init_data(sock, sk);
	sk->sk_destruct = fr_sock_destruct;

	if (sk->sk_prot->init)
		err = sk->sk_prot->init(sk);

	if (err) {
		/* release sk on errors */
		sock_orphan(sk);
		sock_put(sk);
	}


 errout:
	module_put(p->prot->owner);
	return err;
}


/*
 * af_flexray notifier to create/remove FleRay netdevice specific structs
 */
static int fr_notifier(struct notifier_block *nb, unsigned long msg,
			void *data)
{
#if 0
	struct net_device *dev = (struct net_device *)data;
	struct dev_rcv_lists *d;



	if (!net_eq(dev_net(dev), &init_net))
		return NOTIFY_DONE;

	if (dev->type != ARPHRD_CAN)
		return NOTIFY_DONE;

	switch (msg) {

	case NETDEV_REGISTER:

		/*
		 * create new dev_rcv_lists for this device
		 *
		 * N.B. zeroing the struct is the correct initialization
		 * for the embedded hlist_head structs.
		 * Another list type, e.g. list_head, would require
		 * explicit initialization.
		 */

		d = kzalloc(sizeof(*d), GFP_KERNEL);
		if (!d) {
			printk(KERN_ERR
			       "can: allocation of receive list failed\n");
			return NOTIFY_DONE;
		}
		d->dev = dev;

		spin_lock(&can_rcvlists_lock);
		hlist_add_head_rcu(&d->list, &can_rx_dev_list);
		spin_unlock(&can_rcvlists_lock);

		break;

	case NETDEV_UNREGISTER:
		spin_lock(&can_rcvlists_lock);

		d = find_dev_rcv_lists(dev);
		if (d) {
			if (d->entries) {
				d->remove_on_zero_entries = 1;
				d = NULL;
			} else
				hlist_del_rcu(&d->list);
		} else
			printk(KERN_ERR "flexray: notifier: receive list not "
			       "found for dev %s\n", dev->name);

		spin_unlock(&can_rcvlists_lock);

		if (d)
			call_rcu(&d->rcu, can_rx_delete_device);

		break;
	}

	return NOTIFY_DONE;
#endif
	return NOTIFY_DONE;
}


static int fr_queue_frame(struct sock *sk, struct sk_buff *skb,
		struct net_device *dev,
		struct packet_type *pt, struct net_device *orig_dev)
{
	#if 0
	struct ec_cb *eb = (struct ec_cb *)&skb->cb;
	struct sockaddr_fr *sec = (struct sockaddr_ec *)&eb->sec;
	memset(sec, 0, sizeof(struct sockaddr_ec));
	sec->sec_family = AF_ECONET;
	sec->type = ECTYPE_PACKET_RECEIVED;
	sec->port = port;
	sec->cb = cb;
	sec->addr.net = net;
	sec->addr.station = stn;
	#endif
	return sock_queue_rcv_skb(sk, skb);
	
}

/**
 * Called when data has been received for the FlexRay protocol
 */
static int fr_rcv(struct sk_buff *skb, struct net_device *dev,
		   struct packet_type *pt, struct net_device *orig_dev)
{
	//struct flexray_frame *f = (struct flexray_frame *) skb->data;
	//struct dev_rcv_lists *d = NULL;
	struct hlist_node *node;
	struct sock *sk;
	struct sk_buff *clone;
	
	//printk(KERN_INFO "fr_rcv\n");
	
	rcu_read_lock();

	/* deliver packet to all sockets */
	sk_for_each(sk, node, &fr_sklist) {
		clone = skb_clone(skb, GFP_ATOMIC);
		if(clone == NULL) {
			printk(KERN_ERR "flexray: out of memory\n");
			goto out;	
		}
		if(fr_queue_frame(sk, clone, dev, pt, orig_dev)) {
			// error, free cloned socket
			kfree_skb(clone);
		}
	}
	// TODO: handle sockets that listen only on specific device?!
out:
	rcu_read_unlock();
	
	
	/* free the skbuff allocated by the netdevice driver */
	kfree_skb(skb);
	
	return 0;
}


/**
 * Transmit a FlexRay frame, this is called e.g. by raw_sendmsg
 * 
 * @param skb pointer to socket buffer with FlexRay frame in data section
 * @return
 *  0 on success
 *  -ENETDOWN when the selected interface is down
 *  -ENOBUFS on full driver queue (see net_xmit_errno())
 *  -ENOMEM when local loopback failed at calling skb_clone()
 *  -EPERM when trying to send on a non-FlexRay interface
 *  -EINVAL when the skb->data does not contain a valid FlexRay frame
 */
int fr_send(struct sk_buff *skb)
{
	struct sk_buff *newskb = NULL;
	struct flexray4linux_frame *f = (struct flexray4linux_frame *)skb->data;
	int err;

	if (skb->len != sizeof(struct flexray4linux_frame) || f->length > 127) {
		kfree_skb(skb);
		return -EINVAL;
	}

	if (skb->dev->type != ARPHRD_FLEXRAY) {
		kfree_skb(skb);
		return -EPERM;
	}

	if (!(skb->dev->flags & IFF_UP)) {
		kfree_skb(skb);
		return -ENETDOWN;
	}

	skb->protocol = htons(ETH_P_FLEXRAY);
	skb_reset_network_header(skb);
	skb_reset_transport_header(skb);
	skb->pkt_type = PACKET_HOST;
	
	/* send to netdevice */
	err = dev_queue_xmit(skb);
	if (err > 0)
		err = net_xmit_errno(err);

	if (err) {
		if (newskb)
			kfree_skb(newskb);
		return err;
	}

	if (newskb)
		netif_rx(newskb);

	/* update statistics */
	//can_stats.tx_frames++;
	//can_stats.tx_frames_delta++;

	return 0;
}
EXPORT_SYMBOL(fr_send);

int fr_rx_register(struct sock *sk)
{
	return fr_insert_socket(&fr_sklist, sk);
}
EXPORT_SYMBOL(fr_rx_register);

int fr_rx_unregister(struct sock *sk)
{
	return fr_remove_socket(&fr_sklist, sk);
}
EXPORT_SYMBOL(fr_rx_unregister);

static struct packet_type fr_packet __read_mostly = {
	.type = __constant_htons(ETH_P_FLEXRAY),
	/* NULL means enable all devices */
	.dev  = NULL,
	/* the function to call when frames are received */
	.func = fr_rcv,
};

static struct net_proto_family fr_family_ops __read_mostly = {
	.family = PF_FLEXRAY,
	.create = fr_create,
	.owner  = THIS_MODULE,
};

/* notifier block for netdevice event */
static struct notifier_block fr_netdev_notifier __read_mostly = {
	.notifier_call = fr_notifier,
};

static __init int fr_init(void)
{
	printk(KERN_INFO "flexray: Initializing FlexRay4Linux protocol stack\n");

	//rcv_cache = kmem_cache_create("can_receiver", sizeof(struct receiver),
		//		      0, 0, NULL);
	//if (!rcv_cache)
		//return -ENOMEM;

	/*
	 * Insert can_rx_alldev_list for reception on all devices.
	 * This struct is zero initialized which is correct for the
	 * embedded hlist heads, the dev pointer, and the entries counter.
	 */
#if 0
	spin_lock(&can_rcvlists_lock);
	hlist_add_head_rcu(&can_rx_alldev_list.list, &can_rx_dev_list);
	spin_unlock(&can_rcvlists_lock);
#endif

#if 0
	if (stats_timer) {
		/* the statistics are updated every second (timer triggered) */
		setup_timer(&can_stattimer, can_stat_update, 0);
		mod_timer(&can_stattimer, round_jiffies(jiffies + HZ));
	} else
		can_stattimer.function = NULL;
#endif

	//can_init_proc();

	/* protocol register */
	sock_register(&fr_family_ops);
	register_netdevice_notifier(&fr_netdev_notifier);
	dev_add_pack(&fr_packet);

	return 0;
}

static __exit void fr_exit(void)
{
#if 0
	struct dev_rcv_lists *d;
	struct hlist_node *n, *next;
#endif

#if 0
	if (stats_timer)
		del_timer(&can_stattimer);
#endif

	//fr_remove_proc();

	/* protocol unregister */
	dev_remove_pack(&fr_packet);
	unregister_netdevice_notifier(&fr_netdev_notifier);
	sock_unregister(PF_FLEXRAY);

	/* remove fr_rx_dev_list */
	/*
	spin_lock(&fr_rcvlists_lock);
	hlist_del(&fr_rx_alldev_list.list);
	hlist_for_each_entry_safe(d, n, next, &fr_rx_dev_list, list) {
		hlist_del(&d->list);
		kfree(d);
	}
	*/
	//spin_unlock(&fr_rcvlists_lock);

	//kmem_cache_destroy(rcv_cache);
	
	printk(KERN_INFO "flexray: Successfully unloaded.\n");
}

module_init(fr_init);
module_exit(fr_exit);
