/*
 * You received this file as part of
 * FlexRay4Linux - FlexRay protocol stack for Linux
 * 
 * This file contains:
 *
 * - RAW sockets for the FlexRay protocol family
 *
 * Copyright:
 * 
 * - (C) 2009-2011 Michael Arndt <flexray4linux@scriptkiller.de>,
 *   Robotics Research Lab, University of Kaiserslautern
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT OF THIRD PARTY RIGHTS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) AND AUTHOR(S) BE LIABLE FOR ANY
 * CLAIM, OR ANY SPECIAL INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * ALL LIABILITY, INCLUDING LIABILITY FOR INFRINGEMENT OF ANY PATENTS,
 * COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS, RELATING TO USE OF THIS
 * SOFTWARE IS DISCLAIMED.
 *
 * The provided data structures and external interfaces from this code
 * are not restricted to be used by modules with a GPL compatible license.
 * 
 */


#include <linux/module.h>

#include <linux/types.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/skbuff.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <net/sock.h>
#include <asm/ioctls.h>
#include <linux/kmod.h>
#include <linux/if_arp.h>

#include <linux/flexray.h>
#include "af_flexray.h"

#define VERSION "0.1"

MODULE_DESCRIPTION("FlexRay protocol stack for Linux: RAW sockets");
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Michael Arndt <flexray4linux@scriptkiller.de>");
MODULE_VERSION(VERSION);
MODULE_ALIAS_NET_PF_PROTO(PF_FLEXRAY, FR_RAW);

/**
 * This is the description of a raw FlexRay socket
 */
struct raw_sock {
	struct sock sk;
	int bound;
	int ifindex;
	struct notifier_block notifier;
	int loopback;
	int recv_own_msgs;
};

/**
 * Cast sock struct to raw_sock struct
 */
static inline struct raw_sock *raw_sk(const struct sock *sk)
{
	return (struct raw_sock *)sk;
}

static int raw_notifier(struct notifier_block *nb,
			unsigned long msg, void *data)
{
#if 0
	struct net_device *dev = (struct net_device *)data;
	struct raw_sock *ro = container_of(nb, struct raw_sock, notifier);
	struct sock *sk = &ro->sk;
#endif

	return NOTIFY_DONE;
}

/**
 * Initialize the given raw socket
 * @param sk the socket
 * @return 0 on success
 */
static int raw_init(struct sock *sk)
{
	struct raw_sock *ro = raw_sk(sk);

	printk(KERN_INFO "flexray: raw_init\n");
	fr_rx_register(sk);

	ro->bound            = 0;
	ro->ifindex          = 0;

	/* set notifier */
	ro->notifier.notifier_call = raw_notifier;

	register_netdevice_notifier(&ro->notifier);

	return 0;
}

/**
 * Release the given raw socket
 * @param sock the socket
 * @return 0 on success
 */
static int raw_release(struct socket *sock)
{
	struct sock *sk = sock->sk;
	struct raw_sock *ro = raw_sk(sk);

	printk(KERN_INFO "flexray: raw_release\n");
	fr_rx_unregister(sk);
	
	unregister_netdevice_notifier(&ro->notifier);

	lock_sock(sk);

	ro->ifindex = 0;
	ro->bound   = 0;

	release_sock(sk);
	sock_put(sk);

	return 0;
}


static int raw_bind(struct socket *sock, struct sockaddr *uaddr, int len)
{
	struct sockaddr_fr *addr = (struct sockaddr_fr *)uaddr;
	struct sock *sk = sock->sk;
	struct raw_sock *ro = raw_sk(sk);
	int ifindex;
	int err = 0;
	int notify_enetdown = 0;

	printk(KERN_INFO "flexray: raw_bind\n");

	if (len < sizeof(*addr))
		return -EINVAL;

	lock_sock(sk);

	// TODO this is important, index is needed ...
	if (addr->fr_ifindex) {
		struct net_device *dev;

		dev = dev_get_by_index(&init_net, addr->fr_ifindex);
		if (!dev) {
			printk(KERN_ERR "flexray: raw_bind dev_get_by_index"
				"failed, index = %d\n", addr->fr_ifindex);
			err = -ENODEV;
			goto out;
		}
		if (dev->type != ARPHRD_FLEXRAY) {
			dev_put(dev);
			err = -ENODEV;
			goto out;
		}
		if (!(dev->flags & IFF_UP))
			notify_enetdown = 1;

		ifindex = dev->ifindex;

		dev_put(dev);
	} else {
		ifindex = 0;
	}
	
	ro->ifindex = ifindex;

out:

	release_sock(sk);

	return err;
}

static unsigned int raw_poll(struct file *file, struct socket *sock,
	poll_table  *wait)
{
	//struct sock *sk = sock->sk;
	//struct dn_scp *scp = DN_SK(sk);
	int mask = datagram_poll(file, sock, wait);

	/*
	if (!skb_queue_empty(&scp->other_receive_queue))
		mask |= POLLRDBAND;
	*/
	
	return mask;
}

static int raw_getname(struct socket *sock, struct sockaddr *uaddr,
		       int *len, int peer)
{
	return -EOPNOTSUPP;
}


static int raw_setsockopt(struct socket *sock, int level, int optname,
			  char __user *optval, unsigned int optlen)
{
	return -EOPNOTSUPP;
}

static int raw_getsockopt(struct socket *sock, int level, int optname,
			  char __user *optval, int __user *optlen)
{
	return -EOPNOTSUPP;
}

static int raw_sendmsg(struct kiocb *iocb, struct socket *sock,
		       struct msghdr *msg, size_t size)
{
	struct sock *sk = sock->sk;
	struct raw_sock *ro = raw_sk(sk);
	struct sk_buff *skb;
	struct net_device *dev;
	int ifindex;
	int err;
	
	//printk(KERN_INFO "flexray: raw_sendmsg\n");

	// TODO: this might go wrong, even in net/can/raw.c
	if (msg->msg_name) {
		struct sockaddr_fr *addr =
			(struct sockaddr_fr *)msg->msg_name;

		if (addr->fr_family != AF_FLEXRAY) {
			printk(KERN_ERR "flexray: fr_family not set to AF_FLEXRAY\n");
			return -EINVAL;
		}

		ifindex = addr->fr_ifindex;
	} else
		ifindex = ro->ifindex;
	
	if (size != sizeof(struct flexray4linux_frame)) {
		printk(KERN_ERR "flexray: sendmsg size missmatch\n"); 
		return -EINVAL;
	}

	dev = dev_get_by_index(&init_net, ifindex);
	if (!dev)
		return -ENXIO;

	skb = sock_alloc_send_skb(sk, size, msg->msg_flags & MSG_DONTWAIT,
				  &err);
	if (!skb) {
		printk(KERN_ERR "flexray: sock_alloc_send_skb failed!\n"); 
		dev_put(dev);
		return err;
	}

	err = memcpy_fromiovec(skb_put(skb, size), msg->msg_iov, size);
	if (err < 0) {
		kfree_skb(skb);
		dev_put(dev);
		return err;
	}
	skb->dev = dev;
	skb->sk  = sk;

	err = fr_send(skb);

	dev_put(dev);

	if (err)
		return err;

	return size;
}

static int raw_recvmsg(struct kiocb *iocb, struct socket *sock,
		       struct msghdr *msg, size_t size, int flags)
{
	struct sock *sk = sock->sk;
	struct sk_buff *skb;
	int err = 0;
	int noblock;

	//printk(KERN_INFO "raw_recvmsg\n");

	noblock =  flags & MSG_DONTWAIT;
	flags   &= ~MSG_DONTWAIT;

	skb = skb_recv_datagram(sk, flags, noblock, &err);
	if (!skb)
		return err;

	if (size < skb->len)
		msg->msg_flags |= MSG_TRUNC;
	else
		size = skb->len;

	err = memcpy_toiovec(msg->msg_iov, skb->data, size);
	if (err < 0) {
		skb_free_datagram(sk, skb);
		return err;
	}

	sock_recv_timestamp(msg, sk, skb);

	if (msg->msg_name) {
		msg->msg_namelen = sizeof(struct sockaddr_fr);
		memcpy(msg->msg_name, skb->cb, msg->msg_namelen);
	}

	skb_free_datagram(sk, skb);

	return size;
}


/**
 * possible operations on FR_RAW sockets
 */
static struct proto_ops raw_ops __read_mostly = {
	.family        = PF_FLEXRAY,
	.release       = raw_release,
	.bind          = raw_bind,
	.connect       = sock_no_connect,
	.socketpair    = sock_no_socketpair,
	.accept        = sock_no_accept,
	.getname       = raw_getname,
	.poll          = raw_poll,
	.ioctl         = NULL,		/* use fr_ioctl() from af_flexray.c */
	.listen        = sock_no_listen,
	.shutdown      = sock_no_shutdown,
	.setsockopt    = raw_setsockopt,
	.getsockopt    = raw_getsockopt,
	.sendmsg       = raw_sendmsg,
	.recvmsg       = raw_recvmsg,
	.mmap          = sock_no_mmap,
	.sendpage      = sock_no_sendpage,
};

static struct proto raw_proto __read_mostly = {
	.name       = "FR_RAW",
	.owner      = THIS_MODULE,
	.obj_size   = sizeof(struct raw_sock),
	.init       = raw_init,
};

/**
 * description of the FR_RAW protocol
 */
static struct fr_proto raw_fr_proto __read_mostly = {
	/* type argument in socket() syscall */
	.type       = SOCK_RAW,
	/* protocol number in socket() syscall */
	.protocol   = FR_RAW,
	/* capability needed to open the socket, or -1 for no restriction */
	.capability = -1,
	/* pointer to struct proto_ops for sock->ops */
	.ops        = &raw_ops,
	/* pointer to struct proto structure */
	.prot       = &raw_proto,
};

/**
 * initialize the FR_RAW module
 */
static __init int raw_module_init(void)
{
	int err;

	printk(KERN_INFO "flexray-raw: raw protocol initializing\n");

	/* try to register the FR_RAW protocol */
	err = fr_proto_register(&raw_fr_proto);
	if (err < 0)
		printk(KERN_ERR "flexray-raw: registration of raw protocol failed\n");

	return err;
}

/**
 * unload the FR_RAW module
 */
static __exit void raw_module_exit(void)
{
	fr_proto_unregister(&raw_fr_proto);
	printk(KERN_INFO "flexray-raw: raw protocol successfully unloaded\n");
}

module_init(raw_module_init);
module_exit(raw_module_exit);
